// DESCRIPTION OF THE DATABASE FORMAT
//
// THE DATA FILE CONTAINS SEVERAL SECTIONS.  EACH BEGINS WITH A LINE CONTAINING
// A NUMBER IDENTIFYING THE SECTION, AND ENDS WITH A LINE CONTAINING "-1".
//
// SECTION 1: LONG FORM DESCRIPTIONS.  EACH LINE CONTAINS A LOCATION NUMBER,
//	A TAB, AND A LINE OF TEXT.  THE SET OF (NECESSARILY ADJACENT) LINES
//	WHOSE NUMBERS ARE X FORM THE LONG DESCRIPTION OF LOCATION X.
//
// SECTION 2: SHORT FORM DESCRIPTIONS.  SAME FORMAT AS LONG FORM.  NOT ALL
//	PLACES HAVE SHORT DESCRIPTIONS.
//
// SECTION 3: TRAVEL TABLE.  EACH LINE CONTAINS A LOCATION NUMBER (X), A SECOND
//	LOCATION NUMBER (Y), AND A LIST OF MOTION NUMBERS (SEE SECTION 4).
//	EACH MOTION REPRESENTS A VERB WHICH WILL GO TO Y IF CURRENTLY AT X.
//	Y, IN TURN, IS INTERPRETED AS FOLLOWS.  LET M=Y/1000, N=Y MOD 1000.
//		IF N<=300	IT IS THE LOCATION TO GO TO.
//		IF 300<N<=500	N-300 IS USED IN A COMPUTED GOTO TO
//					A SECTION OF SPECIAL CODE.
//		IF N>500	MESSAGE N-500 FROM SECTION 6 IS PRINTED,
//					AND HE STAYS WHEREVER HE IS.
//	MEANWHILE, M SPECIFIES THE CONDITIONS ON THE MOTION.
//		IF M=0		IT'S UNCONDITIONAL.
//		IF 0<M<100	IT IS DONE WITH M% PROBABILITY.
//		IF M=100	UNCONDITIONAL, BUT FORBIDDEN TO DWARVES.
//		IF 100<M<=200	HE MUST BE CARRYING OBJECT M-100.
//		IF 200<M<=300	MUST BE CARRYING OR IN SAME ROOM AS M-200.
//		IF 300<M<=400	PROP(M MOD 100) MUST *NOT* BE 0.
//		IF 400<M<=500	PROP(M MOD 100) MUST *NOT* BE 1.
//		IF 500<M<=600	PROP(M MOD 100) MUST *NOT* BE 2, ETC.
//	IF THE CONDITION (IF ANY) IS NOT MET, THEN THE NEXT *DIFFERENT*
//	"DESTINATION" VALUE IS USED (UNLESS IT FAILS TO MEET *ITS* CONDITIONS,
//	IN WHICH CASE THE NEXT IS FOUND, ETC.).  TYPICALLY, THE NEXT DEST WILL
//	BE FOR ONE OF THE SAME VERBS, SO THAT ITS ONLY USE IS AS THE ALTERNATE
//	DESTINATION FOR THOSE VERBS.  FOR INSTANCE:
//		15	110022	29	31	34	35	23	43
//		15	14	29
//	THIS SAYS THAT, FROM LOC 15, ANY OF THE VERBS 29, 31, ETC., WILL TAKE
//	HIM TO 22 IF HE'S CARRYING OBJECT 10, AND OTHERWISE WILL GO TO 14.
//		11	303008	49
//		11	9	50
//	THIS SAYS THAT, FROM 11, 49 TAKES HIM TO 8 UNLESS PROP(3)=0, IN WHICH
//	CASE HE GOES TO 9.  VERB 50 TAKES HIM TO 9 REGARDLESS OF PROP(3).
//
// SECTION 4: VOCABULARY.  EACH LINE CONTAINS A NUMBER (N), A TAB, AND A
//	FIVE-LETTER WORD.  CALL M=N/1000.  IF M=0, THEN THE WORD IS A MOTION
//	VERB FOR USE IN TRAVELLING (SEE SECTION 3).  ELSE, IF M=1, THE WORD IS
//	AN OBJECT.  ELSE, IF M=2, THE WORD IS AN ACTION VERB (SUCH AS "CARRY"
//	OR "ATTACK").  ELSE, IF M=3, THE WORD IS A SPECIAL CASE VERB (SUCH AS
//	"DIG") AND N MOD 1000 IS AN INDEX INTO SECTION 6.  OBJECTS FROM 50 TO
//	(CURRENTLY, ANYWAY) 79 ARE CONSIDERED TREASURES (FOR PIRATE, CLOSEOUT).
//
// SECTION 5: OBJECT DESCRIPTIONS.  EACH LINE CONTAINS A NUMBER (N), A TAB,
//	AND A MESSAGE.  IF N IS FROM 1 TO 100, THE MESSAGE IS THE "INVENTORY"
//	MESSAGE FOR OBJECT N.  OTHERWISE, N SHOULD BE 000, 100, 200, ETC., AND
//	THE MESSAGE SHOULD BE THE DESCRIPTION OF THE PRECEDING OBJECT WHEN ITS
//	PROP VALUE IS N/100.  THE N/100 IS USED ONLY TO DISTINGUISH MULTIPLE
//	MESSAGES FROM MULTI-LINE MESSAGES; THE PROP INFO ACTUALLY REQUIRES ALL
//	MESSAGES FOR AN OBJECT TO BE PRESENT AND CONSECUTIVE.  PROPERTIES WHICH
//	PRODUCE NO MESSAGE SHOULD BE GIVEN THE MESSAGE ">$<".
//
// SECTION 6: ARBITRARY MESSAGES.  SAME FORMAT AS SECTIONS 1, 2, AND 5, EXCEPT
//	THE NUMBERS BEAR NO RELATION TO ANYTHING (EXCEPT FOR SPECIAL VERBS
//	IN SECTION 4).
//
// SECTION 7: OBJECT LOCATIONS.  EACH LINE CONTAINS AN OBJECT NUMBER AND ITS
//	INITIAL LOCATION (ZERO (OR OMITTED) IF NONE).  IF THE OBJECT IS
//	IMMOVABLE, THE LOCATION IS FOLLOWED BY A "-1".  IF IT HAS TWO LOCATIONS
//	(E.G. THE GRATE) THE FIRST LOCATION IS FOLLOWED WITH THE SECOND, AND
//	THE OBJECT IS ASSUMED TO BE IMMOVABLE.
//
// SECTION 8: ACTION DEFAULTS.  EACH LINE CONTAINS AN "ACTION-VERB" NUMBER AND
//	THE INDEX (IN SECTION 6) OF THE DEFAULT MESSAGE FOR THE VERB.
//
// SECTION 9: LIQUID ASSETS, ETC.  EACH LINE CONTAINS A NUMBER (N) AND UP TO 20
//	LOCATION NUMBERS.  BIT N (WHERE 0 IS THE UNITS BIT) IS SET IN COND(LOC)
//	FOR EACH LOC GIVEN.  THE COND BITS CURRENTLY ASSIGNED ARE:
//		0	LIGHT
//		1	IF BIT 2 IS ON: ON FOR OIL, OFF FOR WATER
//		2	LIQUID ASSET, SEE BIT 1
//		3	PIRATE DOESN'T GO HERE UNLESS FOLLOWING PLAYER
//	OTHER BITS ARE USED TO INDICATE AREAS OF INTEREST TO "HINT" ROUTINES:
//		4	TRYING TO GET INTO CAVE
//		5	TRYING TO CATCH BIRD
//		6	TRYING TO DEAL WITH SNAKE
//		7	LOST IN MAZE
//		8	PONDERING DARK ROOM
//		9	AT WITT'S END
//	COND(LOC) IS SET TO 2, OVERRIDING ALL OTHER BITS, IF LOC HAS FORCED
//	MOTION.
//
// SECTION 10: CLASS MESSAGES.  EACH LINE CONTAINS A NUMBER (N), A TAB, AND A
//	MESSAGE DESCRIBING A CLASSIFICATION OF PLAYER.  THE SCORING SECTION
//	SELECTS THE APPROPRIATE MESSAGE, WHERE EACH MESSAGE IS CONSIDERED TO
//	APPLY TO PLAYERS WHOSE SCORES ARE HIGHER THAN THE PREVIOUS N BUT NOT
//	HIGHER THAN THIS N.  NOTE THAT THESE SCORES PROBABLY CHANGE WITH EVERY
//	MODIFICATION (AND PARTICULARLY EXPANSION) OF THE PROGRAM.
//
// SECTION 11: HINTS.  EACH LINE CONTAINS A HINT NUMBER (CORRESPONDING TO A
//	COND BIT, SEE SECTION 9), THE NUMBER OF TURNS HE MUST BE AT THE RIGHT
//	LOC(S) BEFORE TRIGGERING THE HINT, THE POINTS DEDUCTED FOR TAKING THE
//	HINT, THE MESSAGE NUMBER (SECTION 6) OF THE QUESTION, AND THE MESSAGE
//	NUMBER OF THE HINT.  THESE VALUES ARE STASHED IN THE "HINTS" ARRAY.
//	HNTMAX IS SET TO THE MAX HINT NUMBER (<= HNTSIZ).  NUMBERS 1-3 ARE
//	UNUSABLE SINCE COND BITS ARE OTHERWISE ASSIGNED, SO 2 IS USED TO
//	REMEMBER IF HE'S READ THE CLUE IN THE REPOSITORY, AND 3 IS USED TO
//	REMEMBER WHETHER HE ASKED FOR INSTRUCTIONS (GETS MORE TURNS, BUT LOSES
//	POINTS).
//
// SECTION 12: MAGIC MESSAGES. IDENTICAL TO SECTION 6 EXCEPT PUT IN A SEPARATE
//	SECTION FOR EASIER REFERENCE.  MAGIC MESSAGES ARE USED BY THE STARTUP,
//	MAINTENANCE MODE, AND RELATED ROUTINES.
//
// SECTION 0: END OF DATABASE.
//

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

//
// CURRENT LIMITS:
//     9650 WORDS OF MESSAGE TEXT (LINES, LINSIZ).
//	750 TRAVEL OPTIONS (TRAVEL, TRVSIZ).
//	300 VOCABULARY WORDS (KTAB, ATAB, TABSIZ).
//	150 LOCATIONS (LTEXT, STEXT, KEY, COND, ABB, ATLOC, LOCSIZ).
//	100 OBJECTS (PLAC, PLACE, FIXD, FIXED, LINK (TWICE), PTEXT, PROP).
//	 35 "ACTION" VERBS (ACTSPK, VRBSIZ).
//	205 RANDOM MESSAGES (RTEXT, RTXSIZ).
//	 12 DIFFERENT PLAYER CLASSIFICATIONS (CTEXT, CVAL, CLSMAX).
//	 20 HINTS, LESS 3 (HINTLC, HINTED, HINTS, HNTSIZ).
//	 35 MAGIC MESSAGES (MTEXT, MAGSIZ).
// THERE ARE ALSO LIMITS WHICH CANNOT BE EXCEEDED DUE TO THE STRUCTURE OF
// THE DATABASE.  (E.G., THE VOCABULARY USES N/1000 TO DETERMINE WORD TYPE,
// SO THERE CAN'T BE MORE THAN 1000 WORDS.)  THESE UPPER LIMITS ARE:
//	1000 NON-SYNONYMOUS VOCABULARY WORDS
//	300 LOCATIONS
//	100 OBJECTS

const int LINSIZ=65536;
const int TRVSIZ=750;
const int TABSIZ=300;
const int LOCSIZ=150;
const int OBJSIZ=100;
const int VRBSIZ=35;
const int RTXSIZ=205;
const int CLSMAX=12;
const int HNTSIZ=20;
const int MAGSIZ=35;

char LINES[LINSIZ];
int TRAVEL[TRVSIZ];
int KTAB[TABSIZ];
char ATAB[TABSIZ*5];
int LTEXT[LOCSIZ];
int STEXT[LOCSIZ];
int PTEXT[OBJSIZ];
int PINDX[OBJSIZ*3];
int RTEXT[RTXSIZ];
int CTEXT[CLSMAX];
int CVAL[CLSMAX];
int MTEXT[MAGSIZ];
int KEY[LOCSIZ];
int PLAC[OBJSIZ];
int FIXD[OBJSIZ];
int ACTSPK[35];
int COND[LOCSIZ];

struct {
    int TRNS;
    int PNTS;
    int QSTN;
    int HINT;
} HINTS[HNTSIZ];

int CLSSES;
int HNTMAX;

void BUG(int n);  // forward reference


int read_integer(FILE* fp, int* var)
{
  int c,i;
  char buf[12];

  c = fgetc(fp);  // the first character must be a - or digit
  buf[0] = c;

  for (i=1; i<12; i++) {
    c = fgetc(fp);
    if (!isdigit(c)) break;
    buf[i] = c;
  }
  buf[i] = '\0';
  *var = atoi(buf);
  return c;
}

// Read the database

int LINUSE;
int TRVS;
int PRPUSE;
int TABNDX;

void read_db()
{
  FILE* fp;
  int I;
  int K;
  int LOC;
  int NEWLOC;
  int OBJ;
  int OLDLOC;
  int SECT;
  int TK[21];
  int VERB;

  printf(" Initializing...\n");

// CLEAR OUT THE VARIOUS TEXT-POINTER ARRAYS.  ALL TEXT IS STORED IN ARRAY
// LINES; EACH LINE IS PRECEDED BY A WORD POINTING TO THE NEXT POINTER (I.E.
// THE WORD FOLLOWING THE END OF THE LINE).  THE POINTER IS NEGATIVE IF THIS IS
// FIRST LINE OF A MESSAGE.  THE TEXT-POINTER ARRAYS CONTAIN INDICES OF
// POINTER-WORDS IN LINES.  STEXT(N) IS SHORT DESCRIPTION OF LOCATION N.
// LTEXT(N) IS LONG DESCRIPTION.  PTEXT(N) POINTS TO MESSAGE FOR PROP(N)=0.
// SUCCESSIVE PROP MESSAGES ARE FOUND BY CHASING POINTERS.  RTEXT CONTAINS
// SECTION 6'S STUFF.  CTEXT(N) POINTS TO A PLAYER-CLASS MESSAGE.  MTEXT IS FOR
// SECTION 12.  WE ALSO CLEAR COND.  SEE DESCRIPTION OF SECTION 9 FOR DETAILS.

	for (I=1; I<300; I++) {
	  if (I < OBJSIZ) PTEXT[I]=0;
	  if (I < RTXSIZ) RTEXT[I]=0;
	  if (I < CLSMAX) CTEXT[I]=0;
	  if (I < MAGSIZ) MTEXT[I]=0;
	  if (I < LOCSIZ) {
	    STEXT[I]=0;
	    LTEXT[I]=0;
	    COND[I]=0;
	  }
        }

	fp = fopen("cave.dat", "ra");
	if (fp == NULL) {
	  perror("Unable to open file advent.dat");
          exit(1);
	}
	LINUSE=1;
	PRPUSE=1;
	TRVS=1;
	CLSSES=1;

// START NEW DATA SECTION.  SECT IS THE SECTION NUMBER.

L1002:	fscanf(fp, "%d\n", &SECT);
	OLDLOC=-1;
        switch (SECT) {
	  case 0: goto L1100;
	  case 1: goto L1004;
	  case 2: goto L1004;
	  case 3: goto L1030;
	  case 4: goto L1040;
	  case 5: goto L1010;
	  case 6: goto L1004;
	  case 7: goto L1050;
	  case 8: goto L1060;
	  case 9: goto L1070;
	  case 10: goto L1004;
	  case 11: goto L1080;
	  case 12: goto L1004;
	  default: BUG(9);
	}

// SECTIONS 1, 2, 5, 6, 10, 12.  READ MESSAGES AND SET UP POINTERS.

L1004:	fscanf(fp, "%d\t", &LOC);
	if (LOC==-1) goto L1002;
	if (LOC!=OLDLOC) LINUSE++;
	fgets(&LINES[LINUSE], LINSIZ-LINUSE, fp);
	if (ferror(fp)) {
	  perror("Error while reading file advent.dat");
	  exit(1);
	}
	if (LOC==OLDLOC) {
	  LINUSE += strlen(&LINES[LINUSE]);
	  goto L1004;
	}
	if (SECT==12) goto L1013;
	if (SECT==10) goto L1012;
	if (SECT==6) goto L1011;
	if (SECT==5) goto L1010;
	if (SECT==1) goto L1008;

	STEXT[LOC]=LINUSE;
	goto L1020;

L1008:	LTEXT[LOC]=LINUSE;
	goto L1020;

L1010:	if (LOC > 0 & LOC < OBJSIZ) {
	  PTEXT[LOC]=PRPUSE;
	  PINDX[PRPUSE++]=-LINUSE;
	}
	else {
	  PINDX[PRPUSE++]=LINUSE;
	}
	goto L1020;

L1011:	if (LOC >= RTXSIZ) BUG(6);
	RTEXT[LOC]=LINUSE;
	goto L1020;

L1012:	CTEXT[CLSSES]=LINUSE;
	CVAL[CLSSES]=LOC;
	CLSSES=CLSSES+1;
	goto L1020;

L1013:	if (LOC >= MAGSIZ) BUG(6);
	MTEXT[LOC]=LINUSE;

//L1020: LINUSE=KK+1;
//       LINES(LINUSE)=-1;
//       if (LINUSE+14>LINSIZ) BUG(2);

L1020:	LINUSE += strlen(&LINES[LINUSE]);
	OLDLOC=LOC;
        if (LINUSE+80 >= LINSIZ) BUG(2);
	goto L1004;


// THE STUFF FOR SECTION 3 IS ENCODED HERE.  EACH "FROM-LOCATION" GETS A
// CONTIGUOUS SECTION OF THE "TRAVEL" ARRAY.  EACH ENTRY IN TRAVEL IS
// NEWLOC*1000 + KEYWORD (FROM SECTION 4, MOTION VERBS), AND IS NEGATED IF
// THIS IS THE LAST ENTRY FOR THIS LOCATION.  KEY(N) IS THE INDEX IN TRAVEL
// OF THE FIRST OPTION AT LOCATION N.

L1030:	read_integer(fp, &LOC);
	if (LOC==-1) goto L1002;
	read_integer(fp, &NEWLOC);
	for (int i=1; i<=20; i++) {
          int c;
          if (i == 1 || c == '\t')
	    c = read_integer(fp, &TK[i]);
	  else
	    TK[i] = 0;
        }
	//printf("LOC=%d NEWLOC=%d TK=%d",LOC,NEWLOC,TK[1]);
        //for (int i=2; i<=20; i++) if (TK[i]!=0) printf(",%d", TK[i]);
	//printf("\n");

	if (LOC >= LOCSIZ) BUG(10);
	if (KEY[LOC]!=0) goto L1033;
	KEY[LOC]=TRVS;
	goto L1035;
L1033:	TRAVEL[TRVS-1]=-TRAVEL[TRVS-1];
L1035:	for (int L=1; L<=20; L++) {
	 if (TK[L]==0) goto L1039;
	 TRAVEL[TRVS]=NEWLOC*1000+TK[L];
	 TRVS=TRVS+1;
	 if (TRVS>=TRVSIZ) BUG(3);
        }
L1039:	TRAVEL[TRVS-1]=-TRAVEL[TRVS-1];
	goto L1030;

// HERE WE READ IN THE VOCABULARY.  KTAB(N) IS THE WORD NUMBER, ATAB(N) IS
// THE CORRESPONDING WORD.  THE -1 AT THE END OF SECTION 4 IS LEFT IN KTAB
// AS AN END-MARKER.  THE WORDS ARE GIVEN A MINIMAL HASH TO MAKE READING THE
// CORE-IMAGE HARDER.  NOTE THAT '/7-08' HAD BETTER NOT BE IN THE LIST, SINCE
// IT COULD HASH TO -1.

 L1040:	for (TABNDX=1; TABNDX < TABSIZ; TABNDX++) {
          int c;
          c = read_integer(fp, &KTAB[TABNDX]);
	  if (KTAB[TABNDX]==-1) goto L1002;
	  if (c != '\t') BUG(11);
	  ATAB[TABNDX*5] = fgetc(fp);
	  c = fgetc(fp);
	  ATAB[TABNDX*5+1] = isalpha(c) ? c : '\0';
	  if (isalpha(c)) c = fgetc(fp);
	  ATAB[TABNDX*5+2] = isalpha(c) ? c : '\0';
	  if (isalpha(c)) c = fgetc(fp);
	  ATAB[TABNDX*5+3] = isalpha(c) ? c : '\0';
	  if (isalpha(c)) c = fgetc(fp);
	  ATAB[TABNDX*5+4] = isalpha(c) ? c : '\0';
	  while (c != '\n') c = fgetc(fp);
        }
//L1042:	ATAB(TABNDX)=ATAB(TABNDX).XOR.'PHROG';
	BUG(4);

// READ IN THE INITIAL LOCATIONS FOR EACH OBJECT.  ALSO THE IMMOVABILITY INFO.
// PLA//CONTAINS INITIAL LOCATIONS OF OBJECTS.  FIXD IS -1 FOR IMMOVABLE
// OBJECTS (INCLUDING THE SNAKE), OR = SECOND LOC FOR TWO-PLACED OBJECTS.

L1050:	read_integer(fp, &OBJ);
	if (OBJ==-1) goto L1002;
	if (read_integer(fp, &PLAC[OBJ]) != '\n') {
	  if (read_integer(fp, &FIXD[OBJ]) != '\n') BUG(12);
        }
	goto L1050;

// READ DEFAULT MESSAGE NUMBERS FOR ACTION VERBS, STORE IN ACTSPK.

L1060:	read_integer(fp, &VERB);
	if (VERB==-1) goto L1002;
	if (read_integer(fp, &ACTSPK[VERB]) != '\n') BUG(13);
	goto L1060;

// READ INFO ABOUT AVAILABLE LIQUIDS AND OTHER CONDITIONS, STORE IN COND.

L1070:	read_integer(fp, &K);
	if (K<0) goto L1002;
	for (int i=1; i<=20; i++) {
          int c;
          if (i == 1 || c == '\t')
	    c = read_integer(fp, &TK[i]);
	  else
	    TK[i] = 0;
        }
	for (I=1; I<=20; I++) {
          if (TK[I]==0) goto L1070;
	  LOC=TK[I];
	  if ((COND[LOC] & (1<<K)) != 0) {
	    printf("K=%d LOC=%d COND[LOC]=%d\n", K, LOC, COND[LOC]);
	    BUG(8);
	  }
	  COND[LOC]=COND[LOC] | (1<<K);
        }
	goto L1070;

// READ DATA FOR HINTS.

L1080:	HNTMAX=0;
L1081:	read_integer(fp, &K);
	if (K==-1) goto L1002;
	if (K<0 | K>HNTSIZ) BUG(7);
	if (K>HNTMAX) HNTMAX=K;
	if (read_integer(fp, &HINTS[K].TRNS) != '\t') BUG(14);
	if (read_integer(fp, &HINTS[K].PNTS) != '\t') BUG(14);
	if (read_integer(fp, &HINTS[K].QSTN) != '\t') BUG(14);
	if (read_integer(fp, &HINTS[K].HINT) != '\n') BUG(14);
	goto L1081;

// DONE, REPORT ON AMOUNT OF ARRAYS ACTUALLY USED, TO PERMIT REDUCTIONS.

L1100:  fclose(fp);

	printf(" Table space used:\n");
	printf("   %d OF %d words of messages\n", LINUSE, LINSIZ);

	for (K=LOCSIZ-1; K>0; K--) {
	  if (LTEXT[K]!=0) break;
	}
	printf("   %d OF %d locations\n", K, LOCSIZ);

	printf("   %d OF %d travel options\n", TRVS, TRVSIZ);

	printf("   %d OF %d vocabulary words\n", TABNDX, TABSIZ);

	for (I=1; I<TABNDX; I++) if (KTAB[I]/1000==2) VERB=KTAB[I]-2000;
	printf("   %d OF %d action verbs\n", VERB, VRBSIZ);

	OBJ=0;
	for (I=1; I<OBJSIZ; I++) if (PTEXT[I]!=0) OBJ++;
	printf("   %d OF %d objects\n", OBJ, OBJSIZ);

	for (K=RTXSIZ-1; K>0; K--) {
	  if (RTEXT[K]!=0) break;
	}
	printf("   %d OF %d RTEXT messages\n", K, RTXSIZ);

	printf("   %d OF %d CLASS messages\n", CLSSES, CLSMAX);

	printf("   %d OF %d hints\n", HNTMAX, HNTSIZ);

	for (K=MAGSIZ-1; K>0; K--) {
	  if (MTEXT[K]!=0) break;
	}
	printf("   %d OF %d MAGIC messages\n", K, MAGSIZ);
}


